;; GCC machine description for Motorola 6809
;; Copyright (C) 1989, 2005, 2006, 2007 Free Software Foundation, Inc.
;;
;; Mostly by Brian Dominy (brian@oddchange.com)
;;
;; Based on earlier work by Tom Jones (jones@sal.wisc.edu) and
;; Matthias Doerfel (msdoerfe@informatik.uni-erlangen.de)
;;
;; This file is part of GCC.
;;
;; GCC is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 1, or (at your option)
;; any later version.
;;
;; GCC is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with GCC; see the file COPYING.  If not, write to
;; the Free Software Foundation, 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.
;;

;; General information:
;; --------------------
;; * This backend is mostly a rewrite from earlier (3.1.1 and before)
;; versions.
;;
;; * The 'A' and 'B' registers are treated a single register by the
;; register allocator; hence, the instruction templates assume that
;; both can be modified if either one is available for use.  No
;; attempt is made to split instructions to refer to a particular half
;; of the register.  It is always refer to as the 'D' register, even
;; in QImode (when it will be displayed as 'B').
;;
;; * There is full support for proper branch instruction generation,
;; based on instruction lengths.  However, many instruction patterns
;; are still overloaded to emit lots of real instructions, which can
;; make the length calculation difficult; in those cases, I've tried
;; to be pessimistic and assume the worst-case.
;;
;; * The instruction type attributes are only defined for branch
;; vs. non branch instructions for now, since there is seemingly no
;; reason to define these for other types anyway.
;;
;; * The limited number of total registers presents the greatest
;; challenge.  Some infrastructure for 'soft registers' -- memory
;; locations used to simulate real regs -- is in place, although it
;; is not used at all now.
;;
;; * Likewise, some work to support PIC has been done but it is
;; far from complete and should not actually be used.
;;
;; * All of the 6809 special opcodes, e.g. SWI and SYNC, are defined
;; as UNSPEC instructions, and can be accessed from C code using
;; __builtin_xxxx() style functions.
;;
;; What needs to be done:
;; ----------------------
;; * Shift and rotate patterns are too complicated.  There is a lot of
;; commonality among them all.  Consider expanders for the named
;; insns, and only use define_insn for real 6809 instructions.
;;
;; * Replace remaining instances of (define_peephole) with
;; (define_peephole2), or remove them completely if they are not
;; matching anyway.  Add more peepholes for things actually encountered.
;;
;; * Indexing addressing can lead to crashes in complex functions when
;; register pressure is high.  Only the 'D' register can actually be
;; used as an index register, and its demand by other instructions
;; can sometimes mean that it is impossible to satisfy constraints.
;; Currently, indexing is completely disabled to avoid these types
;; of problems, although code is slightly more inefficient in some
;; working cases.


;;--------------------------------------------------------------------
;;-  Constants
;;--------------------------------------------------------------------

;
; Define constants for hard register numbers.
;
(define_constants [
  (HARD_RSVD1_REGNUM 0)
  (HARD_X_REGNUM 1) (HARD_Y_REGNUM 2) (HARD_U_REGNUM 3)
  (HARD_S_REGNUM 4) (HARD_PC_REGNUM 5) (HARD_D_REGNUM 6)
  (HARD_Z_REGNUM 7)
  (HARD_A_REGNUM 8) (HARD_B_REGNUM 9)
  (HARD_CC_REGNUM 10) (HARD_DP_REGNUM 11)
  (SOFT_FP_REGNUM 12) (SOFT_AP_REGNUM 13) 
  (SOFT_M0_REGNUM 14) (SOFT_M1_REGNUM 15)
  (SOFT_M2_REGNUM 16) (SOFT_M3_REGNUM 17)
])


;
; The range in which a short branch insn can be used.
;
(define_constants [
  (MIN_SHORT_BRANCH_OFFSET -127)
  (MAX_SHORT_BRANCH_OFFSET 128)
])


;
; The lengths of various types of real 6809 instructions.
;
; By default, ordinary insns are 4 bytes long.  This is often not
; right, and the insn patterns below will redefine this to the
; correct value.
;
; Branch instruction lengths (conditional and unconditionals) are
; well known and declared here.  The short insns are used when the
; offset is within the range declared above (between MIN_SHORT
; and MAX_SHORT) ; otherwise the long form is used.
;
(define_constants [
  (DEFAULT_INSN_LENGTH 4)
  (SHORT_CBRANCH_LENGTH 2)
  (LONG_CBRANCH_LENGTH 4)
  (SHORT_BRANCH_LENGTH 2)
  (LONG_BRANCH_LENGTH 3)
])


;
; Constants for insn cycle counts.
; Note that these counts all assume 1-byte opcodes.  2-byte
; opcodes require 1 extra cycles for fetching the extra byte.
;
(define_constants [
  ;; The default insn length, when it cannot be calculated.
  ;; Take a conservative approach and estimate high.
  (DEFAULT_INSN_CYCLES 10)

  ;; Cycle counts for ALU and load operations.
  (ALU_INHERENT_CYCLES 2)
  (ALU_IMMED_CYCLES 2)
  (ALU_DIRECT_CYCLES 4)
  (ALU_INDEXED_BASE_CYCLES 4)
  (ALU_EXTENDED_CYCLES 5)

  ;; If an ALU operation is on a 16-bit register (D), then
  ;; add this number of cycles to the total count.
  (ALU_16BIT_CYCLES 2)

  ;; A load of a 16-bit register incurs this extra amount.
  (LOAD_16BIT_CYCLES 1)

  ;; Cycle counts for memory-only operations (bit shifts, clear, test)
  (MEM_DIRECT_CYCLES 6)
  (MEM_INDEXED_BASE_CYCLES 6)
  (MEM_EXTENDED_CYCLES 7)

  ;; Cycle count for any reg-reg transfer (regardless of size)
  (EXG_CYCLES 8)
  (TFR_CYCLES 6)

  ;; Cycle count for a condition code update (andcc/orcc)
  (CC_CYCLES 3)

  (JMP_DIRECT_CYCLES 3)
  (JMP_INDEXED_BASE_CYCLES 3)
  (JMP_EXTENDED_CYCLES 4)

  (JSR_DIRECT_CYCLES 7)
  (JSR_INDEXED_BASE_CYCLES 7)
  (JSR_EXTENDED_CYCLES 8)

  (LEA_BASE_CYCLES 4)

  ;; Cycle count for a psh/pul operations.  Add to this the
  ;; total number of bytes moved for the correct count.
  (PSH_PUL_CYCLES 5)

  ;; Miscellaneous cycle counts
  (CWAI_CYCLES 20)
  (MUL_CYCLES 11)
  (NOP_CYCLES 2)
  (RTI_CYCLES 15)
  (RTS_CYCLES 5)
  (SWI_CYCLES 20)
  (SYNC_CYCLES 4)
])


;
; An enumeration of values for each "unspec"; i.e. unspecified
; instruction.  These represent insns that are meaningful on the
; 6809 but which have no intrinsic meaning to GCC itself.
; These insns can be generated explicitly using the __builtin_xxx
; syntax; they are also implicitly generated by the backend
; as needed to implement other insns.
;
(define_constants [
  (UNSPEC_BLOCKAGE 0)
  (UNSPEC_PUSH_RS 1)
  (UNSPEC_POP_RS 2)
  (UNSPEC_SWI 3)
  (UNSPEC_CWAI 4)
  (UNSPEC_ADD_CARRY 5)
  (UNSPEC_SUB_CARRY 6)
  (UNSPEC_SYNC 7)
  (UNSPEC_ADD_DECIMAL 8)
])


;;--------------------------------------------------------------------
;;-  Attributes
;;--------------------------------------------------------------------

;;
;; The type attribute is used to distinguish between different
;; types of branch instructions, so that their lengths can be
;; calculated correctly.
;;
(define_attr "type" "branch,cbranch,unknown" (const_string "unknown"))

;;
;; The length of a branch instruction is calculated based on how
;; far away the branch target is.  Lengths of other insns default
;; to 4.  set_attr is used in instruction templates to specify
;; the length when it is known exactly.  When not sure, err on
;; the high side to avoid compile errors.
;;
(define_attr "length" "" 
  (cond [
    (eq_attr "type" "branch")
    (if_then_else (lt (minus (match_dup 0) (pc))
                      (const_int MIN_SHORT_BRANCH_OFFSET))
      (const_int LONG_BRANCH_LENGTH)
        (if_then_else (gt (minus (match_dup 0) (pc))
                          (const_int MAX_SHORT_BRANCH_OFFSET))
          (const_int LONG_BRANCH_LENGTH)
          (const_int SHORT_BRANCH_LENGTH)))
    (eq_attr "type" "cbranch")
    (if_then_else (lt (minus (match_dup 0) (pc))
                      (const_int MIN_SHORT_BRANCH_OFFSET))
      (const_int LONG_CBRANCH_LENGTH)
        (if_then_else (gt (minus (match_dup 0) (pc))
                          (const_int MAX_SHORT_BRANCH_OFFSET))
          (const_int LONG_CBRANCH_LENGTH)
          (const_int SHORT_CBRANCH_LENGTH)))
  ] (const_int DEFAULT_INSN_LENGTH)))


;;
;; The default attributes for 'asm' statements.
;; The default length is the longest possible single 6809 instruction,
;; which is 4 bytes.  GCC will automatically multiply this by the
;; number of real insns contained in an asm statement.
;;
(define_asm_attributes
  [(set_attr "length" "4")
   (set_attr "type" "unknown")])

;;
;; An attribute for the number of cycles that it takes an instruction
;; to execute.
;;
(define_attr "cycles" "" (const_int DEFAULT_INSN_CYCLES)) 
  


;;--------------------------------------------------------------------
;;- Instruction patterns.  When multiple patterns apply,
;;- the first one in the file is chosen.
;;-
;;- See file "rtl.def" for documentation on define_insn, match_*, et. al.
;;-
;;- Note: NOTICE_UPDATE_CC in m6809.h handles condition code updates
;;- for most instructions.
;;--------------------------------------------------------------------

;;--------------------------------------------------------------------
;;-  Test
;;--------------------------------------------------------------------

(define_insn "tsthi"
  [(set (cc0) (match_operand:HI 0 "register_operand" "r"))]
  ""
  "cmp%0\t#0"
   [(set_attr "length" "4")])


(define_insn "tstqi"
  [(set (cc0) (match_operand:QI 0 "nonimmediate_operand" "q,m,z"))]
  ""
  "@
   tst%0
   tst\t%0
   ;tstz"
   [(set_attr "length" "1,3,0")])

;;--------------------------------------------------------------------
;;- Compare
;;--------------------------------------------------------------------

;;- cmphi for register to memory or register compares
(define_insn "cmphi"
  [(set (cc0)
    (compare 
      (match_operand:HI 0 "general_operand" "da, mi, ??Ud")
      (match_operand:HI 1 "general_operand" "mi, da,  dU")))]
  ""
{
  if ((REG_P (operands[0])) && (REG_P (operands[1]))) {
    output_asm_insn ("pshs\t%1\t;cmphi: R:%1 with R:%0", operands);
    return "cmp%0\t,s++\t;cmphi:";
  }
  if (GET_CODE (operands[0]) == REG)
    return "cmp%0\t%1\t;cmphi:";
  else {
    cc_status.flags |= CC_REVERSED;
    return "cmp%1\t%0\t;cmphi:(R)";
  }
}
  [(set_attr "length" "5,5,7")])


(define_insn "cmpqi"
  [(set (cc0)
    (compare (match_operand:QI 0 "whole_general_operand" "q,q, q,O,mt,K")
    (match_operand:QI 1 "whole_general_operand"          "O,mt,K,q,q, q")))]
  ""
{
    if (Q_REG_P (operands[0]))
    {
      if (operands[1] == const0_rtx)
        return "tst%0\t;cmpqi:(ZERO)";
      else
        return "cmp%0\t%1\t;cmpqi:";
    }
    else
    {
      cc_status.flags |= CC_REVERSED;

      if (operands[0] == const0_rtx)
        return "tst%1\t;cmpqi:(RZERO)";
      else
        return "cmp%1\t%0\t;cmpqi:(R)";
    }
}
   [(set_attr "length" "1,3,2,1,3,2")])


;;--------------------------------------------------------------------
;;-  Move
;;--------------------------------------------------------------------

(define_expand "movhi"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
    (match_operand:HI 1 "general_operand" ""))]
  ""
  "/* One of the ops has to be in a register */
   if (!register_operand (operand0, HImode) &&
       !register_operand (operand1, HImode))
      operands[1] = copy_to_mode_reg (HImode, operand1);
  ")

;;; Try a splitter to handle failure cases where we try to move
;;; an immediate constant (zero usually) directly to memory.
;;; This absolutely requires an intermediate register.
(define_split
  [(set (match_operand:HI 0 "memory_operand" "")
    (match_operand:HI 1 "immediate_operand"  ""))
   (clobber (match_operand:HI 2 "register_operand" ""))]
  ""
  [(set (match_dup 2) (match_dup 1))
   (set (match_dup 0) (match_dup 2))]
  "")


;;; Optimize the move of a byte to the stack using the pshs instruction
;;; instead of a store with pre-increment.
(define_insn "movhi_push"
  [(set (match_operand:HI 0 "push_operand" "=m")
    (match_operand:HI 1 "register_operand" "U"))]
  ""
  "pshs\t%1\t;movhi_push: R:%1"
  [(set_attr "length" "2")])


(define_insn "*movhi_pic_symbolref"
  [(set (match_operand:HI 0 "register_operand" "=a")
    (match_operand:HI 1 "symbolic_operand" ""))]
  "flag_pic"
  "lea%0\t%c1,pcr"
   [(set_attr "length" "4")])


(define_insn "*movhi"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=a,d,a,ad,mu")
    (match_operand:HI 1 "general_operand"          " a,a,d,miu,ad"))]
  ""
  "@
   lea%0\t,%1
   tfr\t%1,%0
   tfr\t%1,%0
   ld%0\t%1
   st%1\t%0"
   [(set_attr "length" "2,2,2,*,*")])


;;; Optimize the move of a byte to the stack using the pshs instruction
;;; instead of a store with post increment.
(define_insn "movqi_push"
  [(set (match_operand:QI 0 "push_operand" "=m")
    (match_operand:QI 1 "register_operand" " q"))]
  ""
  "pshs\t%1\t;movqi_push: R:%1"
  [(set_attr "length" "2")])


;;; Optimize the move of a byte from the stack using the puls instruction
;;; instead of a store with post-decrement.
(define_insn "movqi_pop"
  [(set (match_operand:QI 0 "register_operand" "=q")
    (match_operand:QI 1 "pop_operand" "m"))]
  ""
  "puls\t%0\t;movqi_pop: R:%0"
  [(set_attr "length" "2")])


;;- load low byte of 16-bit data into 8-bit register/memory
(define_insn "*mov_lsb"
  [(set (match_operand:QI 0 "nonimmediate_operand"    "=q,q,q,m,!q")
      (subreg:QI (match_operand:HI 1 "general_operand" "d,m,a,d, U") 1))]
  ""
  "@
   \t;movlsbqihi: D->B
   ld%0\t%L1\t;movlsbqihi: msb:%1 -> R:%0
   tfr\t%1,d\t;movlsbqihi: R:%1 -> R:%0
   stb\t%0\t;movlsbqihi: R:%1 -> %0
   pshs\t%1\t;movlsbqihi: R:%1 -> R:%0\;leas\t1,s\;puls\t%0"
   [(set_attr "length" "0,*,2,*,6")])


;;- load high byte of 16-bit data into 8-bit register/memory
(define_insn "*mov_msb"
  [(set (match_operand:QI 0 "nonimmediate_operand"    "=q,q,q,q,m,!q")
      (subreg:QI (match_operand:HI 1 "general_operand" "d,O,a,m,d, U") 0))]
  ""
  "@
   tfr\ta,b\t;movmsbqihi: D->B
   clr%0\t\t;movmsbqihi: ZERO -> R:%0
   tfr\t%1,d\t;movmsbqihi: R:%1 -> R:%0\;tfr\ta,b
   ld%0\t%L1\t;movmsbqihi: lsb:%1 -> R:%0
   sta\t%0\t;movmsbqihi: R:%1 -> %0
   pshs\t%1\t;movmsbqihi: R:%1 -> R:%0\;puls\t%0\;leas\t1,s" 
  [(set_attr "length" "2,1,4,*,*,6")])


(define_insn "*movqi_boolean"
  [(set (reg:QI HARD_Z_REGNUM)
    (match_operand:QI 0 "general_operand" "q,O,i,m"))]
  ""
  "@
   tst%0
   andcc\t#~4
   orcc\t#4
   tst\t%0")


(define_insn "movqi"
  [(set (match_operand:QI 0 "nonimmediate_operand" "=q,q,tm,q,tm,q,z")
    (match_operand:QI 1 "general_operand"          " q,O,O,tmi,q,z,q"))]
  ""
  "@
   tfr\t%1,%0\t;movqi: R:%1 -> R:%0
   clr%0\t\t;movqi: ZERO -> R:%0
   clr\t%0\t;movqi: ZERO -> R:%0
   ld%0\t%1\t;movqi: %1 -> R:%0
   st%1\t%0\t;movqi: R:%1 -> %0
   tfr\tcc,%0\;and%0\t#4\t;movqi: R%1 -> %0
   tst%0;movqi: R%1 -> %0"
   [(set_attr "length" "2,1,3,*,*,4,1")])


;;--------------------------------------------------------------------
;;-  Swap registers
;;--------------------------------------------------------------------

(define_insn "swapqi"
  [(set (match_operand:QI 0 "register_operand" "+q")
    (match_operand:QI 1 "register_operand" "+q"))
   (set (match_dup 1) (match_dup 0))]
  ""
  "exg\t%1,%0\t;swapqi: R:%1 <-> R:%0"
  [(set_attr "length" "2")
   (set (attr "cycles") (const_int EXG_CYCLES))])


(define_insn "swaphi"
  [(set (match_operand:HI 0 "register_operand" "+r")
    (match_operand:HI 1 "register_operand" "+r"))
   (set (match_dup 1) (match_dup 0))]
  ""
  "exg\t%1,%0\t;swaphi: R:%1 <-> R:%0"
  [(set_attr "length" "2")
   (set (attr "cycles") (const_int EXG_CYCLES))])


;;--------------------------------------------------------------------
;;-  Extension and truncation insns.
;;--------------------------------------------------------------------

(define_insn "extendqihi2"
  [(set (match_operand:HI 0 "register_operand"               "=d,d")
        (sign_extend:HI (match_operand:QI 1 "general_operand" "q,mi")))]
    ""
    "@
     sex\t\t;extendqihi2: R:%1 -> R:%0
     ldb\t%1\t;extendqihi2: %1 -> R:%0\;sex"
  [(set_attr "length" "1,*")])


(define_insn "zero_extendqihi2"
  [(set (match_operand:HI 0 "register_operand"               "=d, d")
        (zero_extend:HI (match_operand:QI 1 "general_operand" "q, mi")))]
  ""
  "@
   clra\t\t;zero_extendqihi: R:%1 -> R:%0
   ldb\t%1\t;zero_extendqihi: %1 -> R:d\;clra"
  [(set_attr "length" "1,*")])


;;--------------------------------------------------------------------
;;- All kinds of add instructions.
;;--------------------------------------------------------------------


;;
;; gcc's automatic version of addsi3 doesn't know about adcb,adca
;; so it is MUCH more inefficient.  Define this one ourselves.
;;
;;; (define_insn "addsi3"
;;;   [(set (match_operand:SI 0 "nonimmediate_operand" "=m")
;;;      (plus:SI (match_operand:SI 1 "general_operand" "m")
;;;         (match_operand:SI 2 "general_operand" "m")))
;;;    (clobber (match_scratch:HI 3 "=d"))]
;;;   ""
;;; {
;;;   rtx xoperands[8];
;;; 
;;;   xoperands[0] = adjust_address (operands[0], HImode, 2);
;;;   xoperands[1] = adjust_address (operands[1], HImode, 2);
;;;   xoperands[2] = adjust_address (operands[2], HImode, 2);
;;;   xoperands[3] = adjust_address (operands[0], HImode, 0);
;;;   xoperands[4] = adjust_address (operands[1], QImode, 0);
;;;   xoperands[5] = adjust_address (operands[1], QImode, 1);
;;;   xoperands[6] = adjust_address (operands[2], HImode, 0);
;;;   xoperands[7] = operands[3];
;;; 
;;;   output_asm_insn (
;;;     "ld%7\t%2"    "\;"
;;;     "add%7\t%1"   "\;"
;;;     "st%7\t%0"    "\;"
;;;     "ld%7\t%6"    "\;"
;;;     "adcb\t%5"   "\;"
;;;     "adca\t%4"   "\;"
;;;     "st%7\t%3", xoperands);
;;;   return "";
;;; }
;;;   [(set_attr "length" "21")])


;;; To be tested later...
;;;;; (define_expand "addsi3"
;;;;;   [(match_operand:SI 0 "nonimmediate_operand" "")
;;;;;    (match_operand:SI 1 "general_operand" "")
;;;;;    (match_operand:SI 2 "general_operand" "")]
;;;;;   ""
;;;;; {
;;;;;   rtx m0_low, m0_mid, m0_high;
;;;;;   rtx m1_low, m1_mid, m1_high;
;;;;;   rtx m2_low, m2_mid, m2_high;
;;;;; 
;;;;;   m0_low = adjust_address (operands[0], HImode, 2);
;;;;;   m1_low = adjust_address (operands[1], HImode, 2);
;;;;;   m2_low = adjust_address (operands[2], HImode, 2);
;;;;;   m0_mid = adjust_address (operands[0], QImode, 1);
;;;;;   m1_mid = adjust_address (operands[1], QImode, 1);
;;;;;   m2_mid = adjust_address (operands[2], QImode, 1);
;;;;;   m0_high = adjust_address (operands[0], QImode, 0);
;;;;;   m1_high = adjust_address (operands[1], QImode, 0);
;;;;;   m2_high = adjust_address (operands[2], QImode, 0);
;;;;; 
;;;;;   gen_addhi3 (m0_low, m1_low, m2_low);
;;;;;   gen_addqi3_carry (m0_mid, m1_mid, m2_mid);
;;;;;   gen_addqi3_carry (m0_high, m1_high, m2_high);
;;;;;   DONE;
;;;;; })



; TODO - this is ugly.  During RTL generation, we don't know what registers
; are available, so the multiple-insn sequences can only be solved
; via 'define_split's during matching.  See andhi3 for an example.
; Keep the constraints with ? modifiers to help reload pick the right
; registers.
;
; The forms are:
; 1. D += D, expand this into a shift instead. (rtx costs should be corrected
; to avoid this even happening...)
; 2. D += U, require U to be pushed to memory.  (Lots of patterns do this
; now, is this a better way?)
; 3. Best choice: 'addd'
; 4. Next best choice: 'lea'
; 5. Hybrid of 3 and 4
; 6. Same as 4, not bad
; 7. BAD, no D register at all
; 8. 'lea', as good as 4.
(define_insn "addhi3"
  [(set (match_operand:HI 0 "register_operand"     "=d, d,  d,  a,?a, a,???T,a")
    (plus:HI(match_operand:HI 1 "register_operand" "%0, 0,  0,  d, 0, a, T, a")
    (match_operand:HI 2 "general_operand"          " 0, !U, mi, a, m, d, T, i")
  ))]
  ""
  "@
   lslb\t\t;addhi: R:%0 += R:%2\;rola\t\t;also R:%0 *= 2
   pshs\t%2\t;addhi: R:%0 += R:%2\;add%0\t,s++
   add%0\t%2\t;addhi: R:%0 += %2
   lea%0\t%1,%2\t;addhi: R:%0 = R:%2 + R:%1
   exg\td,%0\t;addhi: R:%0 = R:%1 + %2\;addd\t%2\;exg\td,%0
   lea%0\t%2,%1\t;addhi: R:%0 = R:%1 + R:%2
   pshs\td\t;addhi bad: R:%0 = R:%1 + R:%2\;tfr\t%1,d\;lea%0\td,%2\;puls\td
   lea%0\t%a2,%1\t;addhi: R:%0 = R:%1 + %2"
   [(set_attr "length" "2,6,*,*,7,*,9,*")])
; TODO - attribute lengths labeled * are overly pessimistic.  Depending on
; the offset of an indexed address, its length will vary.  We assume the
; long case by default.



(define_insn "addqi3_carry"
  [(set (match_operand:QI 0 "nonimmediate_operand" "=q")
    (unspec:QI [
      (match_operand:QI 1 "whole_general_operand" "%0")
      (match_operand:QI 2 "whole_general_operand" "tmi")] UNSPEC_ADD_CARRY))]
  ""
  "adc%0\t%2\t;addqi_carry: R:%0 += %2"
  [(set_attr "length" "*")])


; TBD: specifying 'A' for the first constraint, to force into the A register
; is not working because of the way registers are currently set up.  This will
; take some work to get right.  Thus the second alternative as a backup.
(define_insn "addqi3_decimal"
  [(set (match_operand:QI 0 "nonimmediate_operand" "=A,?q")
    (unspec:QI [
      (match_operand:QI 1 "general_operand"        "%0,0")
      (match_operand:QI 2 "general_operand"        "tmi,tmi")] UNSPEC_ADD_DECIMAL))]
  ""
  "@
   adda\t%2\;daa
   tfr\t%0,a\;adda\t%2\;daa\;tfr\ta,%0"
  [(set_attr "length" "5,9")])


(define_insn "addqi3"
  [(set (match_operand:QI 0 "nonimmediate_operand"       "=q,q,q,tm,tm,q")
    (plus:QI (match_operand:QI 1 "whole_general_operand" "%0,0,0,0,0,0")
    (match_operand:QI 2 "whole_general_operand"          " 0,I,N,I,N,tmi")))]
  ""
  "@
   asl%0\t\t;addqi: R:%0 = R:%0 + R:%0
   inc%0\t\t;addqi: R:%0 ++
   dec%0\t\t;addqi: R:%0 --
   inc\t%0\t;addqi: %0 ++
   dec\t%0\t;addqi: R:%0 --
   add%0\t%2\t;addqi: R:%0 += %2"
  [(set_attr "length" "1,1,1,3,3,*")])


;;--------------------------------------------------------------------
;;- Subtract instructions.
;;--------------------------------------------------------------------

(define_insn "subhi3"
  [(set (match_operand:HI 0 "register_operand"      "=d,  d, a")
    (minus:HI (match_operand:HI 1 "register_operand" "0,  0, 0")
    (match_operand:HI 2 "general_operand"            "mi, ?U,n")))]
  ""
  "@
   sub%0\t%2\t;subhi: R:%0 -= %2
   pshs\t%2\t;subhi: R:%0 -= R:%2\;sub%0\t,s++
   lea%0\t%n2,%1\t;subhi: R:%0 = R:%1 + %n2"
   [(set_attr "length" "*,5,3")])


(define_insn "subqi3_carry"
  [(set (match_operand:QI 0 "register_operand" "=q")
    (unspec:QI [
      (match_operand:QI 1 "whole_general_operand" "%0")
      (match_operand:QI 2 "whole_general_operand" "tmi")] UNSPEC_SUB_CARRY))]
  ""
  "sbc%0\t%2\t;subqi_carry: R:%0 += %2"
  [(set_attr "length" "*")])


(define_insn "subqi3"
  [(set (match_operand:QI 0 "register_operand"            "=q, q, !q, !q,   q")
    (minus:QI (match_operand:QI 1 "whole_register_operand" "0, 0,  I,  tmn, 0")
    (match_operand:QI 2 "whole_general_operand"            "I, mn, 0,  0,   t")))]
  ""
  "@
   dec%0\t\t;subqi: R:%0 -= 1
   sub%0\t%2\t;subqi: R:%0 -= %2
   dec%0\t\t;subqi: R:%0 = 1 - R:%0\;neg%0
   sub%0\t%1\t;subqi: R:%0 = %1 - R:%0\;neg%0
   sub%0\t%2"
   [(set_attr "length" "1,3,2,4,3")])


;;--------------------------------------------------------------------
;;- Multiply instructions.
;;--------------------------------------------------------------------

; TODO - merge these two instructions, using 'extend_operator' to match
; either signed or zero extension.  Everything else is the same.
(define_insn "mulqihi3"
   [(set (match_operand:HI 0 "register_operand" "=d")
      (mult:HI (sign_extend:HI (match_operand:QI 1 "general_operand" "%q"))
      (match_operand:QI 2 "general_operand" "tmK")))]
  ""
  "lda\t%2\t;mulqihi3: R:%0 = R:%1 * %2\;mul"
  [(set_attr "length" "3")])


(define_insn "umulqihi3"
   [(set (match_operand:HI 0 "register_operand" "=d")
    (mult:HI (zero_extend:HI (match_operand:QI 1 "general_operand" "%q"))
    (match_operand:QI 2 "general_operand" "tmK")))]
  ""
  "lda\t%2\t;umulqihi3: R:%0 = R:%1 * %2\;mul"
  [(set_attr "length" "3")])


; Expand a 16x16 multiplication into either a libcall or a shift.
; If the second operand is a small constant, use the above form.
; prefer to use a sequence of shift instructions.  Otherwise, 
; do a libcall.
(define_expand "mulhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
    (mult:HI (match_operand:HI 1 "general_operand" "")
    (match_operand:HI 2 "nonmemory_operand" "")))]
  ""
{
  /* if (GET_CODE (operands[2]) == CONST_INT &&
      CONST_OK_FOR_LETTER_P (INTVAL (operands[2]), 'K'))
  {
     emit_insn (gen_mulqihi3 (operands[0], operands[1], operands[2]));
     DONE;
  } */
  emit_libcall_insns (HImode, "mulhi3", operands, 2);
  DONE;
})


;;--------------------------------------------------------------------
;;- Divide instructions.
;;--------------------------------------------------------------------

(define_expand "divhi3"
  [(set (match_operand:HI 0 "register_operand" "")
    (div:HI (match_operand:HI 1 "register_operand" "")
    (match_operand:HI 2 "register_operand" "")))]
  ""
{
  emit_libcall_insns (HImode, "divhi3", operands, 2);
  DONE;
})


(define_expand "divqi3"
  [(set (match_operand:QI 0 "register_operand" "")
    (div:QI (match_operand:QI 1 "register_operand" "")
    (match_operand:QI 2 "register_operand" "")))]
  ""
{
  emit_libcall_insns (QImode, "divqi3", operands, 2);
  DONE;
})


(define_expand "udivhi3"
  [(set (match_operand:HI 0 "register_operand" "")
     (udiv:HI (match_operand:HI 1 "register_operand" "")
              (match_operand:HI 2 "register_operand" "")))]
  ""
{
  emit_libcall_insns (HImode, "udivhi3", operands, 2);
  DONE;
})


;;--------------------------------------------------------------------
;;- mod
;;--------------------------------------------------------------------

(define_expand "modhi3"
  [(set (match_operand:HI 0 "register_operand" "")
    (mod:HI (match_operand:HI 1 "register_operand" "")
    (match_operand:HI 2 "register_operand" "")))]
  ""
{
  emit_libcall_insns (HImode, "modhi3", operands, 2);
  DONE;
})


(define_expand "modqi3"
  [(set (match_operand:QI 0 "register_operand" "")
    (mod:QI (match_operand:QI 1 "register_operand" "")
    (match_operand:QI 2 "register_operand" "")))]
  ""
{
  emit_libcall_insns (QImode, "modqi3", operands, 2);
  DONE;
})


(define_expand "umodhi3"
  [(set (match_operand:HI 0 "register_operand" "")
    (umod:HI (match_operand:HI 1 "register_operand" "")
    (match_operand:HI 2 "register_operand" "")))]
  ""
{
  emit_libcall_insns (HImode, "umodhi3", operands, 2);
  DONE;
})



;;--------------------------------------------------------------------
;;- and, or, xor common patterns
;;--------------------------------------------------------------------

; Split a bitwise HImode into two QImode instructions, with one of
; the sources in a pushable register.  The register is pushed onto
; the stack and memory pop operands (,s+) are used in the QI forms.
(define_split
  [(set (match_operand:HI 0 "register_operand" "")
    (match_operator:HI 3 "logical_bit_operator"
      [(match_operand:HI 1 "register_operand" "")
       (match_operand:HI 2 "register_operand" "")]))]
  "reload_completed"
  [(set (mem:HI (pre_dec:HI (reg:HI HARD_S_REGNUM))) (match_dup 2))
   (set (reg:QI HARD_A_REGNUM) (match_op_dup:QI 3
         [(reg:QI HARD_A_REGNUM)
          (mem:QI (post_inc:QI (reg:HI HARD_S_REGNUM)))]))
   (set (reg:QI HARD_D_REGNUM) (match_op_dup:QI 3
         [(reg:QI HARD_D_REGNUM)
          (mem:QI (post_inc:QI (reg:HI HARD_S_REGNUM)))]))
   (use (reg:QI HARD_A_REGNUM))]
{
})

; Split a bitwise HImode into two QImode instructions, with one
; of the sources being a (MEM (MEM (...)); i.e. an indirect memory
; reference.  This requires dereferencing the pointer into a
; temporary register (X), which must be saved/restored around the
; compute instructions.
(define_split
  [(set (match_operand:HI 0 "register_operand" "")
    (match_operator:HI 3 "logical_bit_operator"
      [(match_operand:HI 1 "register_operand" "")
       (mem:HI (match_operand:HI 2 "memory_operand" ""))]))]
  "reload_completed"
  [
   (set (mem:HI (pre_dec:HI (reg:HI HARD_S_REGNUM))) (match_dup 4))
   (set (match_dup 4) (match_dup 2))
   (set (match_dup 4) (mem:HI (match_dup 4)))
   (set (reg:QI HARD_A_REGNUM) (match_op_dup:QI 3
         [(reg:QI HARD_A_REGNUM)
          (mem:QI (post_inc:QI (match_dup 4)))]))
   (set (reg:QI HARD_D_REGNUM) (match_op_dup:QI 3
         [(reg:QI HARD_D_REGNUM)
          (mem:QI (post_inc:QI (match_dup 4)))]))
   (use (reg:QI HARD_A_REGNUM))
   (set (match_dup 4) (mem:HI (post_inc:HI (reg:HI HARD_S_REGNUM))))
   ]
{
  /* Use X for a temporary index register */
  operands[4] = gen_rtx_REG (HImode, HARD_X_REGNUM);
})


; Split a bitwise HImode into two QImode instructions.  This is
; the common case.  This handles splitting when neither of the
; above two cases applies.
(define_split
  [(set (match_operand:HI 0 "register_operand" "")
    (match_operator:HI 3 "logical_bit_operator"
      [(match_operand:HI 1 "register_operand" "")
       (match_operand:HI 2 "general_operand" "")]))]
  "reload_completed"
  [(set (reg:QI HARD_A_REGNUM) (match_op_dup:QI 3
      [(reg:QI HARD_A_REGNUM) (match_dup 4)]))
   (set (reg:QI HARD_D_REGNUM) (match_op_dup:QI 3
      [(reg:QI HARD_D_REGNUM) (match_dup 5)]))
   (use (reg:QI HARD_A_REGNUM))]
{
  if (GET_CODE (operands[2]) == CONST_INT)
  {
    operands[4] = gen_rtx_const_high (operands[2]);
    operands[5] = gen_rtx_const_low (operands[2]);
  }
  else if ((GET_CODE (operands[2]) == MEM)
    && (GET_CODE (XEXP (operands[2], 0)) == MEM))
  {
    FAIL;
  }
  else
  {
    operands[4] = gen_highpart (QImode, operands[2]);
    operands[5] = gen_lowpart (QImode, operands[2]);
  }
})

; Below are the specific cases for each of the operators.
; The QImode versions are the simplest and can be implemented
; directly on the hardware.  The HImode cases are all output
; using one of the above splitting techniques.

;;--------------------------------------------------------------------
;;- and
;;--------------------------------------------------------------------

(define_insn "andhi3"
  [(set (match_operand:HI 0 "register_operand" "=d")
    (and:HI (match_operand:HI 1 "register_operand" "%0")
    (match_operand:HI 2 "general_operand" "mnU")))]
  ""
  "#")


(define_insn "andqi3"
  [(set (match_operand:QI 0 "register_operand"           "=q,q,q,qc")
    (and:QI (match_operand:QI 1 "whole_register_operand" "%0,0,0,0")
    (match_operand:QI 2 "whole_general_operand"          " O,N,tm,i")))]
  ""
  "@
   clr%0\t;andqi(ZERO)
   \t;andqi(-1)
   and%0\t%2\t;andqi: R:%0 &= %2
   and%0\t%2\t;andqi: R:%0 &= %2"
   [(set_attr "length" "1,0,3,2")])


;;--------------------------------------------------------------------
;;- or
;;--------------------------------------------------------------------

(define_insn "iorhi3"
  [(set (match_operand:HI 0 "register_operand" "=d")
    (ior:HI (match_operand:HI 1 "register_operand" "%0")
    (match_operand:HI 2 "general_operand" "mnU")))]
  ""
  "#")


(define_insn "iorqi3"
  [(set (match_operand:QI 0 "register_operand"           "=q,q, qc")
    (ior:QI (match_operand:QI 1 "whole_register_operand" "%0,0, 0")
    (match_operand:QI 2 "whole_general_operand"          " O,tm,i")))]
  ""
  "@
   \t;iorqi(ZERO)
   or%0\t%2\t;iorqi: %0 |= %2
   or%0\t%2\t;iorqi: %0 |= %2"
   [(set_attr "length" "0,3,2")])

;;--------------------------------------------------------------------
;;- xor
;;--------------------------------------------------------------------

(define_insn "xorhi3"
  [(set (match_operand:HI 0 "register_operand" "=d")
    (xor:HI (match_operand:HI 1 "register_operand" "%0")
    (match_operand:HI 2 "general_operand" "mnU")))]
  ""
  "#")


(define_insn "xorqi3"
  [(set (match_operand:QI 0 "register_operand"           "=q,q,q,q")
    (xor:QI (match_operand:QI 1 "whole_register_operand" "%0,0,0,0")
    (match_operand:QI 2 "whole_general_operand"          " O,N,tm,i")))]
  ""
  "@
   \t;xorqi(ZERO)
   com%0\t;xorqi(-1)
   eor%0\t%2\t;xorqi: %0 XOR= %2
   eor%0\t%2\t;xorqi: %0 XOR= %2"
   [(set_attr "length" "0,1,3,2")])

;;--------------------------------------------------------------------
;;-  Two's Complements
;;--------------------------------------------------------------------

;(define_expand "neghi2"
;  [(set (match_operand:HI 0 "nonimmediate_operand" "")
;    (neg:HI (match_operand:HI 1 "general_operand" "")))]
;  ""
;{
;  rtx temp;
;  emit_insn (gen_one_cmplhi2 (temp, operands[1]));
;  emit_insn (gen_addhi3 (operands[0], temp, const1_rtx));
;  DONE;
;})

(define_insn "neghi2"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=d,!a")
    (neg:HI (match_operand:HI 1 "general_operand"   "0, 0")))]
  "!TARGET_EXPERIMENT"
{
  if (D_REG_P (operands[0])) {
    return "coma\t\t;neghi: R:%0\;comb\;addd\t#1";
  }
  else {
    output_asm_insn ("exg\td,%0\t;neghi: R:%0 = -(R:%0)", operands);
    return "coma\;comb\;addd\t#1\;exg\td,%0";
  }
}
  [(set_attr "length" "5,9")])


(define_insn "negqi2"
  [(set (match_operand:QI 0 "nonimmediate_operand"    "=q,tm")
    (neg:QI (match_operand:QI 1 "nonimmediate_operand" "0,0")))]
  ""
  "@
   neg%0\t\t;negqi: R:%0
   neg\t%0\t;negqi: %0"
  [(set_attr "length" "1,3")])


;;--------------------------------------------------------------------
;;-  One's Complements
;;--------------------------------------------------------------------

(define_expand "one_cmplhi2"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
    (not:HI (match_operand:HI 1 "general_operand" "")))]
  ""
{
  if (TARGET_EXPERIMENT)
  {
    emit_insn (
      gen_one_cmplqi2 (
        gen_lowpart (QImode, operands[0]),
        gen_lowpart (QImode, operands[1])));
    emit_insn (
      gen_one_cmplqi2 (
        gen_highpart (QImode, operands[0]),
        gen_highpart (QImode, operands[1])));
    DONE;
  }
})

(define_insn "*one_cmplhi2_legacy"
  [(set (match_operand:HI 0 "nonimmediate_operand" "=d,!a,?m")
    (not:HI (match_operand:HI 1 "general_operand"   "0, 0, 0")))]
  "!TARGET_EXPERIMENT"
  "@
   coma\;comb
   exg\td,%0\;coma\;comb\;exg\td,%0
   com\t%0\;com\t%L0"
  [(set_attr "length" "2,6,6")])

(define_insn "one_cmplqi2"
  [(set (match_operand:QI 0 "nonimmediate_operand"    "=q,tm")
    (not:QI (match_operand:QI 1 "nonimmediate_operand" "0,0")))]
  ""
  "@
   com%0\t\t;comqi: R:%0
   com\t%0\t;comqi: %0"
  [(set_attr "length" "1,3")])

;;--------------------------------------------------------------------
;;- Shifts/rotates
;;--------------------------------------------------------------------


(define_expand "ashlhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
    (ashift:HI (match_operand:HI 1 "general_operand" "")
    (match_operand:HI 2 "nonmemory_operand" "")))]
  ""
{
})

(define_expand "ashlqi3"
  [(set (match_operand:QI 0 "nonimmediate_operand" "")
    (ashift:QI (match_operand:QI 1 "general_operand" "")
    (match_operand:QI 2 "general_operand" "")))]
  ""
{
  if (TARGET_WPC && (operand1 == const1_rtx))
  {
    if (!register_operand (operand0, QImode))
      operands[0] = copy_to_mode_reg (QImode, operand0);
  }
})


(define_expand "ashrhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
    (ashiftrt:HI (match_operand:HI 1 "general_operand" "")
    (match_operand:HI 2 "nonmemory_operand" "")))]
  ""
{
})

(define_expand "ashrqi3"
  [(set (match_operand:QI 0 "nonimmediate_operand" "")
    (ashiftrt:QI (match_operand:QI 1 "general_operand" "")
    (match_operand:QI 2 "nonmemory_operand" "")))]
  ""
{
})


(define_expand "lshrhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
    (lshiftrt:HI (match_operand:HI 1 "general_operand" "")
    (match_operand:HI 2 "nonmemory_operand" "")))]
  ""
{
})

(define_expand "lshrqi3"
  [(set (match_operand:QI 0 "nonimmediate_operand" "")
    (lshiftrt:QI (match_operand:QI 1 "general_operand" "")
    (match_operand:QI 2 "nonmemory_operand" "")))]
  ""
{
})

;;; TODO - the rotate instructions as defined here do not have the
;;; semantics that GCC wants.  Our "rotate" corresponds to a 6809
;;; rotate, in which one of the operand bits moves into the carry
;;; register.  GCC wants to see the entire data value rotate.
;;; This requires taking the overflow bit and moving it back into
;;; the data, overwriting the garbage value that was rotated _in_
;;; from the carry before the operation.
(define_expand "rotlhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
    (rotate:HI (match_operand:HI 1 "general_operand" "")
    (match_operand:HI 2 "nonmemory_operand" "")))]
  ""
{
})

(define_expand "rotlqi3"
  [(set (match_operand:QI 0 "nonimmediate_operand" "")
    (rotate:QI (match_operand:QI 1 "general_operand" "")
    (match_operand:QI 2 "nonmemory_operand" "")))]
  ""
{
})


(define_expand "rotrhi3"
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
    (rotatert:HI (match_operand:HI 1 "general_operand" "")
    (match_operand:HI 2 "nonmemory_operand" "")))]
  ""
{
})

(define_expand "rotrqi3"
  [(set (match_operand:QI 0 "nonimmediate_operand" "")
    (rotatert:QI (match_operand:QI 1 "general_operand" "")
    (match_operand:QI 2 "nonmemory_operand" "")))]
  ""
{
})

;;; Begin define_insn patterns

(define_insn ""
  [(set (match_operand:HI 0 "nonimmediate_operand"   "=d,m,d")
    (ashift:HI (match_operand:HI 1 "general_operand" " 0,0,0")
    (match_operand:HI 2 "nonmemory_operand"          " n,n,a")))]
  ""
{
  m6809_output_shift_insn (ASHIFT, operands);
  return "";
}
  [(set_attr "length" "32,96,16")])


(define_insn "ashlqi_mem_1"
  [(set (match_operand:QI 0 "memory_operand" "=m")
    (ashift:QI (match_dup 0) (const_int 1)))]
  ""
  "asl\t%0"
  [(set_attr "length" "3")])

(define_insn "ashlqi_qreg_1"
  [(set (match_operand:QI 0 "register_operand" "=q")
    (ashift:QI (match_operand:QI 1 "register_operand" "0")
    (const_int 1)))]
  ""
  "asl%0"
  [(set_attr "length" "3")])


; On the WPC hardware, there is a shift register that can be used
; to compute (1<<n) efficiently in two instructions.  Note that this
; form only works when using -mint8 though, because C will promote
; to 'int' when doing this operation.  TODO : we need a 16-bit form too.
(define_insn "ashlqi3_wpc"
  [(set (match_operand:QI 0 "nonimmediate_operand" "=q")
    (ashift:QI (match_operand:QI 1 "immediate_operand" "I")
    (match_operand:QI 2 "general_operand" "q")))]
  "TARGET_WPC"
  "st%2\t0x3FF7\;ld%0\t0x3FF7"
  [(set_attr "length" "6")])


(define_insn "ashlqi3_const"
  [(set (match_operand:QI 0 "nonimmediate_operand"       "=q,m")
    (ashift:QI (match_operand:QI 1 "nonimmediate_operand" "0,0")
    (match_operand:QI 2 "immediate_operand"               "n,n")))]
  ""
{
  m6809_output_shift_insn (ASHIFT, operands);
  return "";
}
  [(set_attr "length" "8,24")])



(define_insn ""
  [(set (match_operand:HI 0 "nonimmediate_operand"    "=d,m,d")
    (ashiftrt:HI (match_operand:HI 1 "general_operand" "0,0,0")
    (match_operand:HI 2 "nonmemory_operand"            "n,n,a")))]
  ""
{
  m6809_output_shift_insn (ASHIFTRT, operands);
  return "";
}
  [(set_attr "length" "32,96,16")])


(define_insn ""
  [(set (match_operand:QI 0 "nonimmediate_operand"    "=q,m,q")
    (ashiftrt:QI (match_operand:QI 1 "general_operand" "0,0,0")
    (match_operand:QI 2 "nonmemory_operand"            "n,n,a")))
  ]
  ""
{
  m6809_output_shift_insn (ASHIFTRT, operands);
  return "";
}
  [(set_attr "length" "8,24,16")])


(define_insn ""
  [(set (match_operand:QI 0 "nonimmediate_operand"    "=q,m,q")
    (lshiftrt:QI (match_operand:QI 1 "general_operand" "0,0,0")
    (match_operand:SI 2 "general_operand"              "n,n,a")))]
  ""
{
  m6809_output_shift_insn (LSHIFTRT, operands);
  return "";
}
  [(set_attr "length" "8,24,16")])


(define_insn ""
  [(set (match_operand:HI 0 "nonimmediate_operand" "=dm,d")
    (lshiftrt:HI (match_operand:HI 1 "general_operand" "0,0")
    (match_operand:HI 2 "nonmemory_operand" "n,a")))]
  ""
{
  m6809_output_shift_insn (LSHIFTRT, operands);
  return "";
}
  [(set_attr "length" "64,16")])


;;--------------------------------------------------------------------
;; rotate instructions
;;--------------------------------------------------------------------

(define_insn ""
  [(set (match_operand:HI 0 "nonimmediate_operand"   "=d,m,d")
    (rotate:HI (match_operand:HI 1 "general_operand"  "0,0,0")
    (match_operand:HI 2 "nonmemory_operand"           "n,n,a")))]
  ""
{
  m6809_output_shift_insn (ROTATE, operands);
  return "";
}
  [(set_attr "length" "32,96,16")])


(define_insn ""
  [(set (match_operand:QI 0 "nonimmediate_operand"  "=q,m,q")
    (rotate:QI (match_operand:QI 1 "general_operand" "0,0,0")
    (match_operand:QI 2 "nonmemory_operand"          "n,n,a")))
  ]
  ""
{
  m6809_output_shift_insn (ROTATE, operands);
  return "";
}
  [(set_attr "length" "8,24,16")])


(define_insn ""
  [(set (match_operand:HI 0 "nonimmediate_operand"     "=d,m,d")
    (rotatert:HI (match_operand:HI 1 "general_operand"  "0,0,0")
    (match_operand:HI 2 "nonmemory_operand"             "n,n,a")))]
  ""
{
  m6809_output_shift_insn (ROTATE, operands);
  return "";
}
  [(set_attr "length" "32,96,16")])


(define_insn ""
  [(set (match_operand:QI 0 "nonimmediate_operand"    "=q,m,q")
    (rotatert:QI (match_operand:QI 1 "general_operand" "0,0,0")
    (match_operand:QI 2 "nonmemory_operand"            "n,n,a")))
  ]
  ""
{
  m6809_output_shift_insn (ROTATE, operands);
  return "";
}
  [(set_attr "length" "8,24,16")])


;;--------------------------------------------------------------------
;;-  Jumps and transfers
;;--------------------------------------------------------------------

;; A tablejump operation gives the address in operand 0, with the
;; CODE_LABEL for the table in operand 1.  The 'define_expand'
;; shows the arguments as GCC presents them.  However, operand 0
;; is a complex MEM and so the 'define_insn' is written slightly
;; differently so that the actual offset from the jump table can
;; be used (required for PIC mode, where the insn sequence is more
;; complicated).  Two insn forms are given, one for PIC and one for
;; non-PIC.
(define_expand "tablejump"
  [(parallel [
    (set (pc) (match_operand:HI 0 "" ""))
    (use (label_ref (match_operand 1 "" "")))
    (clobber (match_scratch:HI 2 ""))
    ])]
  ""
{
})

(define_insn ""
  [(parallel [
     (set (pc) 
        (mem:HI 
           (plus:HI (match_operand:HI 0 "register_operand" "a")
                    (label_ref (match_operand 1 "" "")))))
     (use (label_ref (match_dup 1)))
     (clobber (match_scratch:HI 2 ""))
     ])]
  "!flag_pic"
{
  output_asm_insn ("jmp\t[%a1,%0]", operands);
  return "";
}
  [(set_attr "length" "4")])


(define_insn ""
  [(parallel [
     (set (pc) 
        (mem:HI 
           (plus:HI (match_operand:HI 0 "register_operand" "d")
                    (label_ref (match_operand 1 "" "")))))
     (use (label_ref (match_dup 1)))
     (clobber (match_scratch:HI 2 "=&a"))
     ])]
  "flag_pic"
{
  output_asm_insn ("lea%2\t%a1,pcr", operands);
  output_asm_insn ("ld%0\t%0,%2", operands);
  output_asm_insn ("jmp\t%0,%2", operands);
  return "";
}
  [(set_attr "length" "8")])



(define_insn "indirect_jump"
  [(set (pc)
    (match_operand:HI 0 "register_operand" "a"))]
  ""
  "jmp\t,%0"
  [(set_attr "length" "3")])


(define_insn "jump"
  [(set (pc) (label_ref (match_operand 0 "" "")))]
  ""
{
  return output_branch_insn ( LABEL_REF, operands, get_attr_length (insn));
}
  [(set (attr "type") (const_string "branch"))])


; Output assembly for a condition branch instruction.
(define_insn "*cond_branch"
  [(set (pc)
    (if_then_else
      (match_operator 1 "comparison_operator" [(cc0) (const_int 0)])
        (label_ref (match_operand 0 "" "")) (pc)))]
  ""
{
  return output_branch_insn ( GET_CODE(operands[1]), 
    operands, get_attr_length (insn));
}
  [(set (attr "type") (const_string "cbranch"))])


; Similar to above, but for a condition branch instruction that
; had its operands reversed at some point.
(define_insn "*cond_branch_reverse"
  [(set (pc) 
    (if_then_else 
      (match_operator 1 "comparison_operator" [(cc0) (const_int 0)])
      (pc) (label_ref (match_operand 0 "" ""))))]
  ""
{
  return output_branch_insn ( reverse_condition (GET_CODE(operands[1])),
    operands, get_attr_length (insn));
}
  [(set (attr "type") (const_string "cbranch"))])


;; These are only used for RTL generation; the above patterns will
;; generate the assembly code.
(define_insn "beq" [(set (pc) (if_then_else (eq (cc0) (const_int 0))
  (label_ref (match_operand 0 "" "")) (pc)))] "" "")

(define_insn "bne" [(set (pc) (if_then_else (ne (cc0) (const_int 0))
      (label_ref (match_operand 0 "" "")) (pc)))] "" "")

(define_insn "bgt" [(set (pc) (if_then_else (gt (cc0) (const_int 0))
      (label_ref (match_operand 0 "" "")) (pc)))] "" "")

(define_insn "bgtu" [(set (pc) (if_then_else (gtu (cc0) (const_int 0))
      (label_ref (match_operand 0 "" "")) (pc)))] "" "")

(define_insn "blt" [(set (pc) (if_then_else (lt (cc0) (const_int 0))
      (label_ref (match_operand 0 "" "")) (pc)))] "" "")

(define_insn "bltu" [(set (pc) (if_then_else (ltu (cc0) (const_int 0))
      (label_ref (match_operand 0 "" "")) (pc)))] "" "")

(define_insn "bge" [(set (pc) (if_then_else (ge (cc0) (const_int 0))
      (label_ref (match_operand 0 "" "")) (pc)))] "" "")

(define_insn "bgeu" [(set (pc) (if_then_else (geu (cc0) (const_int 0))
      (label_ref (match_operand 0 "" "")) (pc)))] "" "")

(define_insn "ble" [(set (pc) (if_then_else (le (cc0) (const_int 0))
      (label_ref (match_operand 0 "" "")) (pc)))] "" "")

(define_insn "bleu" [(set (pc) (if_then_else (leu (cc0) (const_int 0))
      (label_ref (match_operand 0 "" "")) (pc)))] "" "")


;;--------------------------------------------------------------------
;;-  Calls
;;--------------------------------------------------------------------

;; Generate a call instruction for a function that does not
;; return a value.  The expander is used during RTL generation.
;; The instructions below are used during matching; only one
;; of them will be used, depending on the type of function
;; being called.  The different conditions are:
;;
;;    1) far_functionp - is this a far function?  Those need
;;    to be output as indirect calls through a far-function
;;    handler.
;;
;;    2) noreturn_functionp - if the function does not return,
;;    we can use a 'jmp' instead of a 'jsr' to call it.
;;
;;    3) is PIC mode enabled?  If so, we'll always use
;;    relative calls (lbsr or lbra).
;;
;; Note: not all combinations are fully supported, especially
;; relating to PIC.
;;
;; TODO - can we use the 'bsr' instruction anywhere?

(define_expand "call"
  [(call (match_operand:HI 0 "memory_operand" "")
    (match_operand:HI 1 "general_operand" ""))]
  ""
  "")

(define_insn "*call_nopic_far"
  [(call (match_operand:HI 0 "memory_operand" "m")
    (match_operand:HI 1 "general_operand" "g"))]
  "far_functionp (operands[0])"
{
  output_far_call_insn (operands, 0);
  return "";
}
  [(set_attr "length" "6")])


; PIC forms come first, and should only match
; (MEM (SYMBOL_REF)).  Other MEM forms are treated as usual.
(define_insn "*call_pic"
  [(call (mem:HI (match_operand:HI 0 "symbolic_operand" ""))
    (match_operand:HI 1 "general_operand" "g"))]
  "flag_pic && !noreturn_functionp (operands[0])"
  "lbsr\t%C0"
  [(set_attr "length" "4")])


(define_insn "*call_nopic"
  [(call (match_operand:HI 0 "memory_operand" "m")
    (match_operand:HI 1 "general_operand" "g"))]
  "!noreturn_functionp (operands[0])"
  "jsr\t%0\t;CALL: (VOIDmode) (%a1 bytes)"
  [(set_attr "length" "3")
   (set (attr "cycles") (const_int JSR_EXTENDED_CYCLES))])


(define_insn "*call_noreturn_pic"
  [(call (mem:HI (match_operand:HI 0 "symbolic_operand" ""))
    (match_operand:HI 1 "general_operand" "g"))]
  "flag_pic && noreturn_functionp (operands[0])"
  "lbra\t%C0"
  [(set_attr "length" "4")])


(define_insn "*call_noreturn_nopic"
  [(call (match_operand:HI 0 "memory_operand" "m")
    (match_operand:HI 1 "general_operand" "g"))]
  "noreturn_functionp (operands[0])"
  "jmp\t%0\t;CALL: (VOIDmode) (%a1 bytes)"
  [(set_attr "length" "3")])


;;
;; Same as above, but for functions that do return a value.
;;
(define_expand "call_value"
  [(set (match_operand 0 "" "")
    (call (match_operand:HI 1 "memory_operand" "")
    (match_operand:HI 2 "general_operand" "")))]
  ""
  "")


(define_insn "*call_value_far"
  [(set (match_operand 0 "" "=gz")
    (call (match_operand:HI 1 "memory_operand" "m")
    (match_operand:HI 2 "general_operand" "g")))]
  "far_functionp (operands[1])"
{
  output_far_call_insn (operands, 1);
  return "";
}
  [(set_attr "length" "6")])


(define_insn "*call_value_pic"
  [(set (match_operand 0 "" "=gz")
    (call (mem:HI (match_operand:HI 1 "symbolic_operand" ""))
    (match_operand:HI 2 "general_operand" "g")))]
  "flag_pic"
  "lbsr\t%C1"
  [(set_attr "length" "4")])


(define_insn "*call_value_nopic"
  [(set (match_operand 0 "" "=gz")
    (call (match_operand:HI 1 "memory_operand" "m")
    (match_operand:HI 2 "general_operand" "g")))]
  ""
  "jsr\t%1"
  [(set_attr "length" "3")
   (set (attr "cycles") (const_int JSR_EXTENDED_CYCLES))])



;;
;; How to generate an untyped call.
;;
(define_expand "untyped_call"
  [(parallel [(call (match_operand 0 "" "")
        (const_int 0))
      (match_operand 1 "" "")
      (match_operand 2 "" "")])]
  ""
{
  int i;

  emit_call_insn (GEN_CALL (operands[0], const0_rtx, NULL, const0_rtx));
  for (i=0; i < XVECLEN (operands[2], 0); i++)
  {
    rtx set = XVECEXP (operands[2], 0, i);
    emit_move_insn (SET_DEST (set), SET_SRC (set));
  }
  emit_insn (gen_blockage ());
  DONE;
})

;;--------------------------------------------------------------------
;;-  Function Entry and Exit
;;--------------------------------------------------------------------

;; On entry to a function, the stack frame looks as follows:
;; - return address (pushed by the caller)
;; - saved registers
;; - local variable storage
;;
;; If the function does not modify the stack after that, then
;; any of these can be accessed directly as an offset from
;; STACK_POINTER_REGNUM.  Otherwise, a frame pointer is required.
;; In that case, the prologue must also initialize HARD_FRAME_POINTER_REGNUM
;; and all references to the stack frame will use that as a base instead.
;;
(define_expand "prologue"
  [(const_int 0)]
  "prologue_epilogue_required ()"
{
  emit_prologue_insns ();
  DONE;
})


;; The function epilogue does exactly the reverse of the prologue,
;; deallocating local variable space, restoring saved registers,
;; and returning.
;;
;; For the 6809, the return may be 'rti' if the function was
;; declared as an interrupt function, but is normally 'rts'.
;;
;; Also, as an optimization, the register restore and the 'rts'
;; can be combined into a single instruction, by adding 'PC' to the
;; list of registers to be restored.  This is only done if there are
;; any saved registers, as 'rts' is more efficient by itself.
;;
(define_expand "epilogue"
  [(const_int 0)]
  "prologue_epilogue_required ()"
{
  emit_epilogue_insns ();
  DONE;
})


;; The inline return instruction
;;(define_insn "return"
;;  [(return)]
;;  "reload_completed && !prologue_epilogue_required (1)"
;;  "rts\t;inline return"
;;  [(set_attr "length" "1")
;;   (set (attr "cycles") (const_int RTS_CYCLES))])


;; The RTS instruction
(define_insn "return_from_epilogue"
  [(return)]
  "reload_completed && !m6809_current_function_has_type_attr_p (\"interrupt\")"
  "rts\t;epilogue return"
  [(set_attr "length" "1")
   (set (attr "cycles") (const_int RTS_CYCLES))])


;; The RTI instruction
(define_insn "return_from_interrupt_epilogue"
  [(return)]
  "reload_completed && m6809_current_function_has_type_attr_p (\"interrupt\")"
  "rti\t;return from interrupt"
  [(set_attr "length" "1")
   (set (attr "cycles") (const_int RTI_CYCLES))])


;;--------------------------------------------------------------------
;;-  Unspecified instructions
;;--------------------------------------------------------------------

;; An instruction that has the effect of an unspec_volatile, but
;; which doesn't require emitting any assembly code.
(define_insn "blockage"
  [(unspec_volatile [(const_int 0)] UNSPEC_BLOCKAGE)]
  ""
  ""
  [(set_attr "length" "0")
   (set (attr "cycles") (const_int 0))])


;; Say how to push multiple registers onto the stack, using
;; the 6809 'pshs' instruction.  The operand is a regset
;; specifying which registers to push.
;;
;; The operand mode is not given intentionally, so as to allow
;; any possible integer mode for the regset.
;;
;; See below for a peephole that can combine consecutive push
;; instructions that qualify for merging.
(define_insn "register_push"
  [(use (reg:HI HARD_S_REGNUM))
    (unspec_volatile 
      [(match_operand 0 "immediate_operand" "")] UNSPEC_PUSH_RS)
    (clobber (reg:HI HARD_S_REGNUM))]
  ""
  "pshs\t%R0\t;push register set"
  [(set_attr "length" "2")
   (set (attr "cycles") (const_int PSH_PUL_CYCLES))])


;; Say how to pop multiple registers from the stack, using
;; the 6809 'puls' instruction.  The operand is the register
;; bitset value.
(define_insn "register_pop"
  [(use (reg:HI HARD_S_REGNUM))
    (unspec_volatile 
      [(match_operand 0 "immediate_operand" "")] UNSPEC_POP_RS)
    (clobber (reg:HI HARD_S_REGNUM))]
  ""
  "puls\t%R0\t;pop register set"
  [(set_attr "length" "2")
   (set (attr "cycles") (const_int PSH_PUL_CYCLES))])


(define_insn "m6809_swi"
  [(unspec_volatile 
    [(match_operand:QI 0 "immediate_operand" "I,n")] UNSPEC_SWI)]
  ""
  "@
   swi
   swi%c0"
  [(set_attr "length" "1,2")
   (set (attr "cycles") (const_int SWI_CYCLES))])


;; Generate the CWAI instruction
(define_insn "m6809_cwai"
  [(unspec_volatile 
    [(match_operand:QI 0 "immediate_operand" "")] UNSPEC_CWAI)]
  ""
  "cwai\t%0\t;wait for interrupt"
  [(set_attr "length" "2")
   (set (attr "cycles") (const_int CWAI_CYCLES))])


;; Generate the SYNC instruction
(define_insn "m6809_sync"
  [(unspec_volatile [(const_int 0)] UNSPEC_SYNC)]
  ""
  "sync\t;sync to interrupt"
  [(set_attr "length" "1")
   (set (attr "cycles") (const_int SYNC_CYCLES))])


;; Generate the NOP instruction
(define_insn "nop"
  [(const_int 0)]
  ""
  "nop"
   [(set_attr "length" "1")
   (set (attr "cycles") (const_int NOP_CYCLES))])


;;--------------------------------------------------------------------
;;- Peepholes
;;--------------------------------------------------------------------

;;; For a void function, merge "leas 1,s; rts" into "puls a,pc".
;;; It corrupts 'a', but who cares.  It's both shorter and faster.
;;; Similar could be done for any value up to 4, since D and X can
;;; be safely corrupted in those cases.
(define_peephole2
  [(set (reg:HI HARD_S_REGNUM)
        (plus:HI (reg:HI HARD_S_REGNUM) (const_int 1)))
   (return)]
  "reload_completed
   && !m6809_current_function_has_type_attr_p (\"interrupt\")
   && m6809_current_function_is_void ()"
  [(use (reg:HI HARD_S_REGNUM))
   (unspec_volatile [(match_dup 0)] UNSPEC_POP_RS)
   (clobber (reg:HI HARD_S_REGNUM))]
{
  operands[0] = gen_rtx_CONST_INT (QImode, A_REGBIT | PC_REGBIT);
})


;;; Replace "clra; aslb; rola" with "aslb; adca #00".
;;; Same length, but 2 cycles faster.
;;; TODO - this actually won't match currently because the aslb; rola
;;; sequence is generated at output time today.
;(define_peephole2
;  [(set (reg:QI HARD_A_REGNUM) (const_int 0))
;   (set (reg:QI HARD_D_REGNUM)
;        (ashift:QI (reg:QI HARD_D_REGNUM) (const_int 1)))
;   (set (reg:QI HARD_A_REGNUM)
;        (rotate:QI (reg:QI HARD_A_REGNUM) (const_int 1)))]
;  ""
;  []
;  "")


;;; Optimize 'leas -1,s' followed by 'stb ,s'.  This can happen if the
;;; function prologue needs to allocate stack space and 'b' is placed
;;; into that local right away.  Combine the stack allocation with the
;;; store using preincrement mode.
(define_peephole2
  [(set (reg:HI HARD_S_REGNUM)
        (plus:HI (reg:HI HARD_S_REGNUM) (const_int -1)))
   (set (mem:QI (reg:HI HARD_S_REGNUM))
        (match_operand:QI 0 "register_operand" ""))]
  ""
  [(set (mem:QI (pre_dec:HI (reg:HI HARD_S_REGNUM))) (match_dup 0))]
  "")


;;; Same as above, but for a 'clr ,s' that follows the prologue.
(define_peephole2
  [(set (reg:HI HARD_S_REGNUM) (plus:HI (reg:HI HARD_S_REGNUM) (const_int -1)))
   (set (mem:QI (reg:HI HARD_S_REGNUM)) (const_int 0))]
  ""
  [(set (mem:QI (pre_dec:HI (reg:HI HARD_S_REGNUM))) (const_int 0))]
  "")


;;; Merge two consecutive push instructions into a single register_push.
(define_peephole2
  [(set (match_operand 0 "push_operand" "")
    (match_operand 1 "register_operand" ""))
   (set (match_operand 2 "push_operand" "")
    (match_operand 3 "register_operand" ""))]
  "reload_completed 
   && m6809_can_merge_pushpop_p (UNSPEC_PUSH_RS, 1 << REGNO (operands[1]), 1 << REGNO (operands[3]))"
  [(parallel [
    (use (reg:HI HARD_S_REGNUM))
    (unspec_volatile [(match_dup 4)] UNSPEC_PUSH_RS)
    (clobber (reg:HI HARD_S_REGNUM))])]
{
  operands[4] = gen_rtx_CONST_INT (QImode, 
    (1 << REGNO (operands[1])) | (1 << REGNO (operands[3])));
})


;;; Convert 'stX ,--s' into a push instruction.  Use the regset
;;; notation, so that it may be combined with an adjacent regset.
;;; TBD - this doesn't compile some code cleanly.
;(define_peephole2
;  [(set (mem:HI (pre_dec:HI (reg:HI HARD_S_REGNUM)))
;        (reg:HI HARD_X_REGNUM))]
;  "reload_completed"
;  [(parallel [
;    (use (reg:HI HARD_S_REGNUM))
;    (unspec_volatile [(match_dup 0)] UNSPEC_PUSH_RS)
;    (clobber (reg:HI HARD_S_REGNUM))])]
;{
;  operands[0] = gen_rtx_CONST_INT (QImode, X_REGBIT);
;})


;;;
;;; q = (q+1)/2 can be optimized as "lsrb; adcb".  This also
;;; won't overflow when q=0xFF.
;;;
(define_peephole
  [(set (reg:QI HARD_D_REGNUM)
    (lshiftrt:QI (plus:HI (match_dup 0) (const_int 1)) (const_int 1)))]
  ""
  "lsrb\;adcb\t#0; peephole"
  [(set_attr "length" "2")])


;;
;; Optimize the case of following a register store with a test
;; of reg or mem just moved.
;;
(define_peephole
  [(set (match_operand:HI 0 "memory_operand" "=m")
  (match_operand:HI 1 "register_operand" "r"))
   (set (cc0) (match_operand:HI 2 "general_operand" "g"))]
  "operands[2] == operands[0] || operands[2] == operands[1]"
  "st%1\t%0\t;movhi: R:%1 -> %0 w/ implied test of %2"
  [(set_attr "length" "4")])


;;    
;; Optimize a pair of SET instructions in which the second insn
;; is the reverse of the first one.  I.e.
;;
;; A = B
;;        ---->  A = B
;; B = A
;;
;; The second insn is redundant.  Define two patterns, one for QI, one for HI.
;; But don't do this if either is VOLATILE.
;;
(define_peephole2
  [(set (match_operand:HI 0 "nonimmediate_operand" "")
        (match_operand:HI 1 "general_operand" ""))
  (set (match_dup 1) (match_dup 0))]
  "!MEM_VOLATILE_P (operands[0]) && !MEM_VOLATILE_P (operands[1])"
  [(set (match_dup 0) (match_dup 1))]
  "")

(define_peephole2
  [(set (match_operand:QI 0 "nonimmediate_operand" "")
        (match_operand:QI 1 "general_operand" ""))
  (set (match_dup 1) (match_dup 0))]
  "!MEM_VOLATILE_P (operands[0]) && !MEM_VOLATILE_P (operands[1])"
  [(set (match_dup 0) (match_dup 1))]
  "")


;;
;; Optimize the sum of an 8-bit and 16-bit using the 'abx' instruction
;; if B and X can be used.
;;
(define_peephole
  [(set (reg:HI HARD_D_REGNUM)
    (zero_extend:HI (match_operand:QI 0 "general_operand" "m,q")))
   (set (reg:HI HARD_X_REGNUM)
    (plus:HI (reg:HI HARD_D_REGNUM) (reg:HI HARD_X_REGNUM)))]
  ""
  "@
   ldb\t%0\;abx
   abx"
  [(set_attr "length" "4,1")])


;;
;; Optimize a simple function call followed by a return to use a "jmp"
;; instruction rather than a "jsr".  The function must take no args and
;; must not return anything.  The called function should not be an
;; interrupt handler; this is not checked though.  Also far functions
;; cannot benefit from this -- there should be a separate version
;; that handles far functions correctly TODO.
;;
;; TODO - is the requirement for no args really valid?  Seems like
;; this would not need to be checked.
(define_peephole
  [(call (match_operand:HI 0 "general_operand" "")
         (match_operand:HI 1 "general_operand" ""))
   (return)
  ]
  "!noreturn_functionp (operands[0]) 
     && !far_functionp (operands[0]) 
     && GET_CODE (operands[1]) == CONST_INT
     && INTVAL(operands[1]) == 0"
  "jmp\t%0\t;jsr/jmp peephole"
  [(set_attr "length" "3")])


;;
;; Work around a compiler bug that generates bad code when copying
;; between 32-bit memory addresses after a libcall.  The problem seen is 
;; that the source is MEM (REG X), but X is used as the reload register.
;; The second half of the copy therefore fails.
;;
;; The solution is to switch the reload register to D, since that is guaranteed
;; not to be in use right after a libcall.
;;
(define_peephole2
  [(set (reg:HI HARD_X_REGNUM) (mem:HI (reg:HI HARD_X_REGNUM)))
   (set (match_operand:HI 0 "general_operand" "") (reg:HI HARD_X_REGNUM))
   (set (reg:HI HARD_X_REGNUM)
      (mem:HI (plus:HI (reg:HI HARD_X_REGNUM) (const_int 2))))
   (set (match_operand:HI 1 "general_operand" "") (reg:HI HARD_X_REGNUM))]
  "reload_completed"
  [(set (reg:HI HARD_D_REGNUM) (mem:HI (reg:HI HARD_X_REGNUM)))
   (set (match_dup 0) (reg:HI HARD_D_REGNUM))
   (set (reg:HI HARD_X_REGNUM)
      (mem:HI (plus:HI (reg:HI HARD_X_REGNUM) (const_int 2))))
   (set (match_dup 1) (reg:HI HARD_X_REGNUM))]
  "")


;; Optimize two consecutive constant QI push operations.
;; Ordinarily these comes out as LDB/STB/LDB/STB, but we can
;; coalesce the constants into a single LDD/STD.
(define_peephole
  [
    (set (match_operand:QI 0 "register_operand" "=q")
      (match_operand:QI 1 "immediate_operand" "K"))
    (set (match_operand:QI 2 "push_operand" "=m")
      (match_operand:QI 3 "register_operand" "0"))
    (set (match_operand:QI 4 "register_operand" "=0")
      (match_operand:QI 5 "immediate_operand" "K"))
    (set (match_operand:QI 6 "push_operand" "=m")
      (match_operand:QI 7 "register_operand" "0"))
  ]
  ""
{
  rtx xoperands[1];

  xoperands[0] = gen_rtx_CONST_INT (HImode, 
    (INTVAL (operands[5]) << 8) + INTVAL (operands[1]));

  output_asm_insn ("ldd\t%0\t;peephole: two constant byte pushes", xoperands);
  output_asm_insn ("std\t,--s", xoperands);
  return "";
}
  [(set_attr "length" "6")])


;; Turn "and then test" into a "bit test" operation.
;; This is the most used peephople.
;; TODO - op 2 could be a mem
(define_peephole
  [(set (match_operand:QI 0 "register_operand" "=q")
    (and:QI (match_operand:QI 1 "register_operand" "0")
      (match_operand:QI 2 "immediate_operand" "i")))
   (set (cc0) (match_dup 0))]
  ""
  "bit%0\t%2\t;peephole: tstqi(andqi): R:%0"
  [(set_attr "length" "3")])


;; Turn a "decrement, then test" sequence into just a "decrement".
;; The test can be omitted, since it is implicitly done.
(define_peephole2
  [(set (match_operand:QI 0 "nonimmediate_operand" "")
    (plus:QI (match_operand:QI 1 "whole_general_operand" "")
    (match_operand:QI 2 "immediate_operand" "")))
   (set (cc0) (match_dup 0))]
  ""
  [(set (match_dup 0) (plus:QI (match_dup 1) (match_dup 2)))]
  "")


;; Merge an indexed register increment with a previous usage.
;; This is usually done automatically, but not always
;; The 'use' should be optional; in all cases where this has been
;; seen, it is required though.
(define_peephole2
  [(set (match_operand:QI 0 "register_operand" "")
     (mem:QI (match_operand:HI 1 "register_operand" "")))
   (use (match_dup 0))
   (set (match_dup 1) (plus:HI (match_dup 1) (const_int 1)))]
  ""
  [(set (match_dup 0) (mem:QI (post_inc:HI (match_dup 1))))
   (use (match_dup 0))]
  "")


;;; Merge "ldX MEM; ldX ,X" into a single instruction using
;;; the indirect mode.
(define_peephole2
  [(set (reg:HI HARD_X_REGNUM)
    (mem:HI (match_operand:HI 0 "general_operand" "")))
   (set (reg:HI HARD_X_REGNUM) (mem:HI (reg:HI HARD_X_REGNUM)))]
  "reload_completed"
  [(set (reg:HI HARD_X_REGNUM)
    (mem:HI (mem:HI (match_operand:HI 0 "general_operand" ""))))]
  "")


;;; Reorder a store followed by a unary operation on that memory
;;; so that the unary is performed and then the store.  Consider
;;; a binary shift operation, which will be decomposed into
;;; identical single shifts, also.
(define_peephole2
  [(set (match_operand:QI 0 "memory_operand" "")
        (match_operand:QI 1 "register_operand" ""))
  (set (match_dup 0)
       (ashift:QI (match_dup 0) (match_operand:QI 2 "immediate_operand" "")))]
  "reload_completed"
  [(set (match_dup 1)
        (ashift:QI (match_dup 1) (match_operand:QI 2 "immediate_operand" "")))
  (set (match_dup 0) (match_dup 1))]
  "")


;;; Replace "andb #N; cmpb #N; bhi" with "andb #N", if it can be proven
;;; that the branch can never occur because of the limited range of B.
;;; N must be a power of two for this to make sense.  This helps with
;;; the default cases of switch statements on a value (x & N).
(define_peephole2
  [(set (match_operand:QI 0 "register_operand" "")
    (and:QI (match_dup 0) (match_operand:QI 1 "immediate_operand" "")))
   (set (cc0)
    (compare (match_dup 0) (match_dup 1)))
   (set (pc) (if_then_else (gtu (cc0) (const_int 0)) (match_operand 2 "" "") (match_operand 3 "" "")))
   ]
  "reload_completed && power_of_two_p (INTVAL (operands[1]) + 1)"
  [(set (match_dup 0) (and:QI (match_dup 0) (match_dup 1)))]
  "")



;;- mode:emacs-lisp
;;- comment-start: ";;- "
;;- eval: (set-syntax-table (copy-sequence (syntax-table)))
;;- eval: (modify-syntax-entry ?[ "(]")
;;- eval: (modify-syntax-entry ?] ")[")
;;- eval: (modify-syntax-entry ?{ "(}")
;;- eval: (modify-syntax-entry ?} "){")
;-; vim: set ts=2:
;-; vim: set expandtab:
;-; vim: set filetype=lisp: 
;;- End:
